/*-
 * Copyright (c) 2015 Weongyo Jeong <weongyo@3rocks.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

if (!String.prototype.format) {
  String.prototype.format = function () {
    var args = arguments;

    return this.replace(/{(\d+)}/g, function (match, number) {
      return typeof args[number] != 'undefined'
        ? args[number]
        : match
        ;
    });
  };
}

var gPacConfig = '                              \
function FindProxyForURL(url, host) {           \
    host = host.toLowerCase();                  \
    if (isPlainHostName(host) ||                \
	isInNetEx(host, "127.0.0.0/8") ||             \
	isInNetEx(host, "10.0.0.0/8") ||              \
	isInNetEx(host, "172.16.0.0/12") ||           \
	isInNetEx(host, "192.168.0.0/16") ||          \
	isInNetEx(host, "fc00::/7") ||                \
	isInNetEx(host, "fe80::/10") ||               \
	shExpMatch(url, "http://mudfish.net/*") ||    \
	shExpMatch(url, "https://mudfish.net/*")) {   \
	    return "DIRECT";                          \
    }                                           \
    return "{0}";                               \
}';

var gStaticNodes = null;
var gStaticNodeIndex = -1;
var gStaticNodePicked = null;
var gStaticNodeSid = -1;
var gStaticNodeConnected = false;
var gStaticNodeUseSecureConnection = false;
var gDomainDumpEnabled = false;
var gDomainDumpHostnames = new Map();

var gUsername = "";
var gPassword = "";

function setProxyDirectMode() {
  chrome.proxy.settings.set({ value: { mode: "direct" } }, function () {
  });
}

function domain_dump_listener(requestDetails) {
  var url = new URL(requestDetails.url);
  if (url.protocol == "http:" || url.protocol == "https:") {
    gDomainDumpHostnames.set(url.hostname, true);
  }
  return { cancel: false };
}

function domain_dump_enable() {
  if (gDomainDumpEnabled) {
    return;
  }
  gDomainDumpEnabled = true;
  gDomainDumpHostnames = new Map();
  chrome.webRequest.onBeforeRequest.addListener(
    domain_dump_listener,
    { urls: ["<all_urls>"] },
    ['blocking']
  );
}

function domain_dump_disable() {
  if (!gDomainDumpEnabled) {
    return;
  }
  gDomainDumpEnabled = false;
  chrome.webRequest.onBeforeRequest.removeListener(domain_dump_listener);
}

chrome.runtime.onMessage.addListener(function (request, sender, sendResponse) {
  if (request.type == "CONNECT") {
    var staticnode = null;
    var i, sid;

    sid = parseInt(request.sid);
    for (i = 0; i < gStaticNodes.length; i++) {
      if (gStaticNodes[i].sid == sid) {
        staticnode = gStaticNodes[i];
        gStaticNodePicked = staticnode;
        break;
      }
    }
    if (staticnode == null || staticnode == undefined) {
      sendResponse({
        error_code: 503,
        error_msg: "No static node found"
      });
      return;
    }
    var use_secure_connection = false;
    if (request.use_secure_connection) {
      dst = "HTTPS " + staticnode.hostname + ":18082";
      use_secure_connection = true;
    } else {
      dst = "PROXY " + staticnode.ip + ":18080";
    }
    var proxyConfig = {
      mode: "pac_script",
      pacScript: {
        data: gPacConfig.format(dst)
      }
    };
    chrome.proxy.settings.set({
      value: proxyConfig
    }, function () {
      gStaticNodeSid = sid;
      gStaticNodeConnected = true;
      gStaticNodeUseSecureConnection = use_secure_connection;
    });
    sendResponse({ error_code: 0 });
  } else if (request.type == "CLEAR_DUMP_DOMAIN") {
    gDomainDumpHostnames = new Map();
    sendResponse({ error_code: 0 });
  } else if (request.type == "DISCONNECT") {
    gStaticNodeSid = -1;
    gStaticNodeConnected = false;
    gStaticNodeUseSecureConnection = false;
    setProxyDirectMode();
    sendResponse({ error_code: 0 });
  } else if (request.type == "SYNC_STATICNODES") {
    fetchStaticNodes_manual(sendResponse);
    /*
     * This need to let chrome know that this will be async.
     */
    return true;
  } else if (request.type == "GET_STATICNODES") {
    if (gStaticNodes == null) {
      sendResponse({ error_code: 503, staticnodes: null });
      return;
    }
    sendResponse({ staticnodes: gStaticNodes });
  } else if (request.type == "GET_SERVER") {
    var staticnode = gStaticNodePicked;

    if (staticnode == undefined || staticnode == null) {
      sendResponse({ location: "Auto Selection" });
      return;
    }
    sendResponse({ location: staticnode.location });
  } else if (request.type == "GET_DUMP_DOMAIN") {
    sendResponse({
      error_code: 0,
      hostnames: Array.from(gDomainDumpHostnames.keys())
    });
  } else if (request.type == "GET_STATUS") {
    sendResponse({
      error_code: 0,
      connected: gStaticNodeConnected,
      sid: gStaticNodeSid,
      use_secure_connection: gStaticNodeUseSecureConnection,
      domain_dump_enabled: gDomainDumpEnabled
    });
  } else if (request.type == "SET_AUTH") {
    gUsername = request.username;
    gPassword = request.password;
    sendResponse({ error_code: 0 });
  } else if (request.type == "SET_DUMP_DOMAIN") {
    if (request.enable) {
      domain_dump_enable();
    } else {
      domain_dump_disable();
    }
    sendResponse({ error_code: 0 });
  } else if (request.type == "SET_PAC_CONFIG") {
    gPacConfig = request.pac_config;
    sendResponse({ error_code: 0 });
  } else {
    console.log("TODO");
    console.log(request.type);
    sendResponse({ error_code: 500 });
  }
});

function fetchStaticNodes_success_callback(data) {

  gStaticNodes = data.staticnodes;
  if (data.default_sid > 0) {
    for (i = 0; i < data.staticnodes.length; i++) {
      if (data.staticnodes[i].sid == data.default_sid)
        break;
    }
    gStaticNodePicked = data.staticnodes[i];
    if (gStaticNodePicked != undefined && gStaticNodePicked != null) {
      gStaticNodeIndex = i;
      return;
    }
  }
  gStaticNodePicked = data.staticnodes[0];
  gStaticNodeIndex = 0;
}

function fetchStaticNodes_error_callback(xhr, op, err) {
  var url;

  console.log("Error: " + xhr.responseText + " Status: " + xhr.status);

  url = chrome.extension.getURL('/html/staticnodes.json');
  $.getJSON(url, function (data) {
    fetchStaticNodes_success_callback(data);
  });
}

function fetchStaticNodes() {
  $.ajax({
    url: "https://mudfish.net/api/staticnodes",
    type: "GET",
    data: {},
    success: fetchStaticNodes_success_callback,
    error: fetchStaticNodes_error_callback
  });
}

function fetchStaticNodes_manual(sendResponse) {
  $.ajax({
    url: "https://mudfish.net/api/staticnodes",
    type: "GET",
    data: {},
    indexValue: { sendResponse: sendResponse },
    success: function (data) {
      fetchStaticNodes_success_callback(data);
      this.indexValue.sendResponse({ error_code: 0 });
    },
    error: function (xhr, op, err) {
      fetchStaticNodes_error_callback(xhr, op, err);
      this.indexValue.sendResponse({
        error_code: 500,
        err_msg: "Failed to sync"
      });
    }
  });
}

$(document).ready(function () {
  chrome.browserAction.setIcon({
    path: "/images/appIcon/default-icon-b.png"
  });
  setProxyDirectMode();
  fetchStaticNodes();

  chrome.webRequest.onAuthRequired.addListener(
    function (details, callbackFn) {
      if (details.isProxy == false) {
        callbackFn({ cancel: false });
        return;
      }
      if (gUsername.length == 0 ||
        gPassword.length == 0) {
        callbackFn({ cancel: false });
        return;
      }
      callbackFn({
        authCredentials: { username: gUsername, password: gPassword }
      });
    },
    { urls: ["<all_urls>"] },
    ['asyncBlocking']
  );
});
